/***************************************************************************
 *                                                                         *
 *  MAXIMUS-CBCS Source Code, Version 1.02                                 *
 *  Copyright 1989, 1990 by Scott J. Dudley.  All rights reserved.         *
 *                                                                         *
 *  USER.BBS conversion utility                                            *
 *                                                                         *
 *  For complete details of the licensing restrictions, please refer to    *
 *  the licence agreement, which is published in its entirety in MAX.C     *
 *  and LICENCE.MAX.                                                       *
 *                                                                         *
 *  USE OF THIS FILE IS SUBJECT TO THE RESTRICTIONS CONTAINED IN THE       *
 *  MAXIMUS-CBCS LICENSING AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF THIS *
 *  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES, OR IF YOU DO NOT HAVE   *
 *  THESE FILES, YOU SHOULD IMMEDIATELY CONTACT THE AUTHOR AT ONE OF THE   *
 *  ADDRESSES LISTED BELOW.  IN NO EVENT SHOULD YOU PROCEED TO USE THIS    *
 *  FILE WITHOUT HAVING ACCEPTED THE TERMS OF THE MAXIMUS-CBCS LICENSING   *
 *  AGREEMENT, OR SUCH OTHER AGREEMENT AS YOU ARE ABLE TO REACH WITH THE   *
 *  AUTHOR.                                                                *
 *                                                                         *
 *  You can contact the author at one of the address listed below:         *
 *                                                                         *
 *  Scott Dudley           FidoNet  1:249/106                              *
 *  777 Downing St.        IMEXnet  89:483/202                             *
 *  Kingston, Ont.         Internet f106.n249.z1.fidonet.org               *
 *  Canada - K7M 5N3       BBS      (613) 389-8315 - HST/14.4K, 24hrs      *
 *                                                                         *
 ***************************************************************************/

#define NOVARS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <mem.h>
#include <time.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "prog.h"
#include "max.h"
#include "cvtusr.h"

char *user_bbs="USER.BBS",
     *user_bak="USER.BAK",
     *user_dat="USER.DAT",
     *user_idx="USER.IDX",
     *users_bbs="USERS.BBS",
     *rec_num="Record %04d\r";

void Format(void);
void Convert_Opus110(void);
void ConvertTo_Opus110(void);
void Convert_Quick(void);
void Convert_Opus103(void);
void Convert_Max(void);
void Check_If_Exist(void);
long CRC(char *s);

byte MaxToOpusPriv(int mpriv);
int OpusToMaxPriv(byte opriv);
struct tm * pascal OpusDate_to_TmDate(struct tm *tmdate,struct _stamp *opusdate);

#ifdef __TURBOC__
long mktime(struct tm *lt);
#endif

#define MAX_id 0x2158414dL /* "MAX!" */
#define OPUS   0x5355504fL


int main(int argc,char *argv[])
{
  int x;

  Hello("CVTUSR","Maximus-CBCS user file conversion utility",VERSION,"1989");

  if (argc < 2)
    Format();

  for (x=1;x < argc;x++)
  {
    if (*argv[x]=='-' || *argv[x]=='/')
    {
      switch(tolower(argv[x][1]))
      {
        case 'o':
          if (eqstri(argv[x]+2,"103"))
            Convert_Opus103();
          else if (eqstri(argv[x]+2,"110"))
            Convert_Opus110();
          else
          {
            printf("\aInvalid command-line parameter:  `%s'!\n",argv[x]);
            Format();
          }
          break;

        case 'x':
          if (eqstri(argv[x]+2,"110"))
            ConvertTo_Opus110();
          else
          {
            printf("\aInvalid command-line parameter:  `%s'!\n",argv[x]);
            Format();
          }
          break;

        case 'q':
          Convert_Quick();
          break;

        case 'm':
          Convert_Max();
          break;

        default:
         printf("\aInvalid command-line parameter:  `%s'!\n",argv[x]);
         Format();
      }
    }
    else
    {
      printf("\aInvalid command-line parameter:  `%s'!\n",argv[x]);
      Format();
    }
  }

  return 0;
}



void Convert_Opus110(void)
{
  struct tm *gmt;
  struct _usr110 usr110;
  struct _usr usr;

  int infile,
      outfile,
      num;

  char temp[PATHLEN];

  Check_If_Exist();

  if ((infile=open(user_dat,O_RDONLY | O_BINARY))==-1)
  {
    printf("Error opening USER.DAT for read!  Aborting...\n");
    exit(1);
  }

  if ((outfile=open(user_bbs,O_CREAT | O_TRUNC | O_WRONLY | O_BINARY,
                    S_IREAD | S_IWRITE))==-1)
  {
    printf("Error opening USER.BBS for write!  Aborting...\n");
    exit(1);
  }

  for (num=0;read(infile,(char *)&usr110,sizeof(struct _usr110))==sizeof(struct _usr110);)
  {
    memset(&usr,'\0',sizeof(struct _usr));

    if (usr110.OPUS_id != OPUS)
    {
      printf("User file grunged:  Invalid record format: 0x%08lx\n",
             usr110.OPUS_id);

      exit(1);
    }

    strcpy(usr.name,usr110.name);
    strcpy(usr.realname,usr110.alias);

    strcpy(usr.city,usr110.city);
    strcpy(usr.pwd,usr110.pwd);
    strcpy(usr.phone,usr110.usrtel);

    usr.times=usr110.times;

    if (usr110.help==HITECH)
      usr.help=HOTFLASH;
    else usr.help=usr110.help;

    usr.tabs=usr110.tabs;
    usr.nulls=(byte)usr110.nulls;

    sprintf(temp,"%d",usr110.msg);
    usr.msg=Set_Area_Name(temp);

    sprintf(temp,"%d",usr110.files);
    usr.files=Set_Area_Name(temp);

    usr.bad_logon=0;
    usr.ibmchars=!(usr110.Bits & NO_IBMCHAR);
    usr.avatar=(usr110.Bits & AVATAR);
    usr.use_lore=(usr110.Bits & USE_LORE);
    usr.more=(usr110.Bits & MORE_PROMPT);
    usr.ansi=(usr110.Bits & ANSI);
    usr.kludge=(usr110.Bits & CONFIG_SET);
    usr.formfeed=(usr110.Bits & FORMFEED);
    usr.key=(char)usr110.ClassLock;
    usr.priv=OpusToMaxPriv(usr110.ClassPriv);
    usr.struct_len=sizeof(struct _usr)/20;

    *usr.ldate='\0';
    
    if (usr110.ludate)
    {
      gmt=localtime(&usr110.ludate);

      if (gmt)
        sprintf(usr.ldate,"%d %s %d %02d:%02d:%02d",gmt->tm_mday,
            months_ab[gmt->tm_mon],gmt->tm_year,
            gmt->tm_hour,gmt->tm_min,gmt->tm_sec);
    }

    usr.time=usr110.time;
    usr.upld=(int)usr110.upld;
    usr.dnld=(int)usr110.dnld;
    usr.dnldl=usr110.dnldl;

    usr.width=usr110.width+2;
    usr.len=usr110.len+2;
    usr.credit=usr110.credit;
    usr.debit=usr110.debit;

    /* If we're undoing a previous conversion */
    if (usr110.extern_id[0]==MAX_id)
    {
      usr.lastread_ptr=((int *)usr110.extern_inf[0])[0];
      usr.bad_logon=usr110.extern_inf[0][2];

      usr.msg=((int *)usr110.extern_inf[0])[1];
      usr.files=((int *)usr110.extern_inf[0])[2];
      strcpy(usr.ldate,&usr110.extern_inf[0][7]);
    }
    else usr.lastread_ptr=num;

    if(!(++num % 32)) printf(rec_num,num);

    write(outfile,(char *)&usr,sizeof(struct _usr));
  }

  printf("Done!      \n");

  close(infile);
  close(outfile);
}


void ConvertTo_Opus110(void)
{
  struct tm lt;

  union stamp_combo laston;
  struct _usr110 usr110;
  struct _usr usr;
  long usrcrc;

  int infile,
      num;

  FILE *outfile,
       *idxfile;

  char string[80];

  int x,
      d1,d3,d4,d5,d6;


  if ((infile=open(user_bbs,O_RDONLY | O_BINARY))==-1)
  {
    printf("Error opening USER.BBS for read!  Aborting...\n");
    exit(1);
  }

  if ((outfile=fopen(user_dat,"wb"))==NULL)
  {
    printf("Error opening USER.DAT for write!  Aborting...\n");
    exit(1);
  }

  if ((idxfile=fopen(user_idx,"wb"))==NULL)
  {
    printf("Error opening USER.IDX for write!  Aborting...\n");
    exit(1);
  }

  for (num=0;read(infile,(char *)&usr,sizeof(struct _usr))==sizeof(struct _usr);)
  {
    memset(&usr110,'\0',sizeof(struct _usr110));

    usr110.OPUS_id=OPUS;

    strcpy(usr110.name,usr.name);
    strcpy(usr110.alias,usr.realname);

    strcpy(usr110.city,usr.city);
    strcpy(usr110.pwd,usr.pwd);
    strcpy(usr110.usrtel,usr.phone);

    usr110.times=usr.times;

    if (usr.help==HOTFLASH)
      usr110.help=HITECH;
    else usr110.help=usr.help;

    usr110.tabs=usr.tabs;
    usr110.nulls=usr.nulls;

    usr110.msg=atoi(Area_Name(usr.msg));
    usr110.files=atoi(Area_Name(usr.msg));

    usr110.Bits=((usr.ibmchars ? 0           : NO_IBMCHAR) |
                 (usr.avatar ? AVATAR        : 0) |
                 (usr.use_lore ? USE_LORE    : 0) |
                 (usr.more ? MORE_PROMPT     : 0) |
                 (usr.ansi ? ANSI            : 0) |
                 (usr.kludge ? CONFIG_SET    : 0) |
                 (usr.formfeed ? FORMFEED    : 0));

    usr110.ClassLock=usr.key;
    usr110.ClassPriv=MaxToOpusPriv(usr.priv);

    usr110.time=usr.time;
    usr110.upld=usr.upld;
    usr110.dnld=usr.dnld;
    usr110.dnldl=usr.dnldl;

    usr110.width=usr.width-2;
    usr110.len=usr.len-2;
    usr110.credit=usr.credit;
    usr110.debit=usr.debit;

    if ((sscanf(usr.ldate,"%d %s %d %d:%d:%d",&d1,string,&d3,&d4,&d5,&d6)) != 6)
    {
      d1=d4=d5=d6=0;
      d3=80;
      *string='\0';
    }

    for (x=0;x < 12;x++)
      if (eqstri(string,months_ab[x]))
      {
        laston.msg_st.date.mo=x+1;
        break;
      }

    if (x==12)          /* Garbage month */
      laston.msg_st.date.mo=1;

    laston.msg_st.date.da=d1;
    laston.msg_st.date.yr=d3-80;
    laston.msg_st.time.hh=d4;
    laston.msg_st.time.mm=d5;
    laston.msg_st.time.ss=d6;

    OpusDate_to_TmDate(&lt,(struct _stamp *)&laston);

    usr110.ludate=mktime(&lt);

    /* We need to steal one of the Opus 1.10 "user" fields to store some   *
     * stuff their structures can't handle...                              */

    usr110.extern_id[0]=MAX_id;
    ((int *)usr110.extern_inf[0])[0]=usr.lastread_ptr;
    ((int *)usr110.extern_inf[0])[1]=usr.msg;
    ((int *)usr110.extern_inf[0])[2]=usr.files;
    usr110.extern_inf[0][6]=(byte)usr.bad_logon;
    strcpy(&usr110.extern_inf[0][7],usr.ldate);

    if(!(++num % 32)) printf(rec_num,num);

    fwrite((char *)&usr110,1,sizeof(struct _usr110),outfile);

    usrcrc=CRC(usr110.name);
    fwrite((char *)&usrcrc,1,sizeof(long),idxfile);
  }

  printf("Done!      \n");

  close(infile);
  fclose(outfile);
  fclose(idxfile);
}


void Convert_Quick(void)
{
  struct _qbbs_user qbu;
  struct _usr usr;

  int infile,
      outfile,
      num;

  Check_If_Exist();

  if ((infile=open(users_bbs,O_RDONLY | O_BINARY))==-1)
  {
    printf("Error opening USERS.BBS for read!  Aborting...\n");
    exit(1);
  }

  if ((outfile=open(user_bbs,O_CREAT | O_TRUNC | O_WRONLY | O_BINARY,
                    S_IREAD | S_IWRITE))==-1)
  {
    printf("Error opening USER.BBS for write!  Aborting...\n");
    exit(1);
  }

  for (num=0;read(infile,(char *)&qbu,sizeof(struct _qbbs_user))==sizeof(struct _qbbs_user);)
  {
    memset(&usr,'\0',sizeof(struct _usr));

    TP_to_Cstr(qbu.name,usr.name);
    TP_to_Cstr(qbu.city,usr.city);
    TP_to_Cstr(qbu.pwd,usr.pwd);

    TP_to_Cstr(qbu.home_phone,usr.phone);

    usr.credit=qbu.credit;
    usr.times=qbu.times;

    usr.upld=qbu.upk;
    usr.dnld=qbu.downk;
    usr.dnldl=qbu.todayk;
    usr.len=(byte)qbu.len;
    usr.width=80;

    usr.help=NOVICE;

    usr.msg=Set_Area_Name("1");
    usr.files=Set_Area_Name("1");

    usr.kludge=FALSE;
    usr.struct_len=sizeof(struct _usr)/20;

    if (qbu.seclvl >= 100)
      usr.priv=SYSOP;
    else if (qbu.seclvl >= 90)
      usr.priv=ASSTSYSOP;
    else if (qbu.seclvl >= 80)
      usr.priv=CLERK;
    else if (qbu.seclvl >= 70)
      usr.priv=EXTRA;
    else if (qbu.seclvl >= 60)
      usr.priv=FAVORED;
    else if (qbu.seclvl >= 50)
      usr.priv=PRIVIL;
    else if (qbu.seclvl >= 40)
      usr.priv=WORTHY;
    else if (qbu.seclvl >= 20)
      usr.priv=NORMAL;
    else usr.priv=DISGRACE;

    if(!(++num % 32)) printf(rec_num,num);

    write(outfile,(char *)&usr,sizeof(struct _usr));
  }

  printf("Done!      \n");

  close(infile);
  close(outfile);
}


void Convert_Opus103(void)
{
  struct _usr usr;
  char temp[50];

  int userfile,
      num;

  if ((userfile=open(user_bbs,O_RDWR | O_BINARY))==-1)
  {
    printf("Error opening USER.BBS for read/write!  Aborting...\n");
    exit(1);
  }

  for (num=0;read(userfile,(char *)&usr,sizeof(struct _usr)) > 0;)
  {
    memset(usr.realname,'\0',sizeof(usr.realname));
    memset(usr.phone,'\0',sizeof(usr.phone));

    /* Reset the kludge bit so that old users get asked the new Maximus    *
     * questions!                                                          */

    usr.kludge=0;

    usr.lastread_ptr=num;
    usr.bad_logon=0;
    usr.key=0;
    usr.struct_len=sizeof(struct _usr)/20;
    usr.flag=0;

    /* Now convert msg/file area numbers, if necessary */

    if (! (isalnum(usr.msg & 0xff) && isalnum(usr.msg >> 8)))
    {
      sprintf(temp,"%d",usr.msg);
      usr.msg=Set_Area_Name(temp);
    }

    if (! (isalnum(usr.files & 0xff) && isalnum(usr.files >> 8)))
    {
      sprintf(temp,"%d",usr.files);
      usr.files=Set_Area_Name(temp);
    }

    if(!(++num % 32)) printf(rec_num,num);

    lseek(userfile,-(long)(sizeof(struct _usr)),SEEK_CUR);
    write(userfile,(char *)&usr,sizeof(struct _usr));
  }

  printf("Done!      \n");

  close(userfile);
}


void Convert_Max(void)
{
  char temp[PATHLEN];

  struct _usr usr;

  int userfile,
      num;

  if ((userfile=open(user_bbs,O_RDWR | O_BINARY))==-1)
  {
    printf("Error opening USER.BBS for read/write!  Aborting...\n");
    exit(1);
  }

  for (num=0;read(userfile,(char *)&usr,sizeof(struct _usr)) > 0;)
  {
    usr.lastread_ptr=num;
    usr.struct_len=sizeof(struct _usr)/20;
    usr.flag=usr.bits=usr.rsvd1=usr.rsvd2=0;

    /* Fix any msg/file area numbers, if necessary */

    if (! (isalnum(usr.msg & 0xff) && isalnum(usr.msg >> 8)))
    {
      sprintf(temp,"%d",usr.msg);
      usr.msg=Set_Area_Name(temp);
    }

    if (! (isalnum(usr.files & 0xff) && isalnum(usr.files >> 8)))
    {
      sprintf(temp,"%d",usr.files);
      usr.files=Set_Area_Name(temp);
    }

    if(!(++num % 32)) printf(rec_num,num);

    lseek(userfile,-(long)(sizeof(struct _usr)),SEEK_CUR);
    write(userfile,(char *)&usr,sizeof(struct _usr));
  }

  printf("Done!      \n");

  close(userfile);
}


void Check_If_Exist(void)
{
  char temp[PATHLEN];

  if (fexist(user_bbs))
  {
    printf("Warning!  USER.BBS already exists.  Overwrite [y,N]? ");
    fgets(temp,80,stdin);

    printf("\n");

    if (toupper(*temp) != 'Y')
    {
      printf("Aborted!\n");
      exit(1);
    }

    if (fexist(user_bak))
      unlink(user_bak);

    rename(user_bbs,user_bak);
  }
}


int OpusToMaxPriv(byte opriv)
{
  switch(opriv)
  {
    case _SYSOP:
      return SYSOP;

    case _ASSTSYSOP:
      return ASSTSYSOP;

    case _CLERK:
      return CLERK;

    case _EXTRA:
      return EXTRA;

    case _FAVORED:
      return FAVORED;

    case _PRIVEL:
      return PRIVIL;

    case _WORTHY:
      return WORTHY;

    case _NORMAL:
      return NORMAL;

    case _LIMITED:
      return LIMITED;

    case _DISGRACE:
      return DISGRACE;

    case _TWIT:
      return TWIT;

    default:
      return HIDDEN;
  }
}


byte MaxToOpusPriv(int mpriv)
{
  switch(mpriv)
  {
    case SYSOP:
      return _SYSOP;

    case ASSTSYSOP:
      return _ASSTSYSOP;

    case CLERK:
      return _CLERK;

    case EXTRA:
      return _EXTRA;

    case FAVORED:
      return _FAVORED;

    case PRIVIL:
      return _PRIVEL;

    case WORTHY:
      return _WORTHY;

    case NORMAL:
      return _NORMAL;

    case LIMITED:
      return _LIMITED;

    case DISGRACE:
      return _DISGRACE;

    case TWIT:
      return _TWIT;

    default:
      return _HIDDEN;
  }
}



void Format(void)
{
  printf("Command-line format:\n\n");

  printf("  CVTUSR [[-x]...]\n\n");

  printf("Valid switches are:\n\n");

  printf("  -q    Convert a QuickBBS-style USERS.BBS file to Max's USER.BBS.\n");
  printf("  -o103 Convert an Opus 1.03-style USER.BBS to Max's USER.BBS\n");
  printf("  -o110 Convert an Opus 1.10-style USER.DAT to Max's USER.BBS\n");
  printf("  -x110 Convert a Max USER.BBS _to_ an Opus 1.10-style USER.DAT.\n");
  printf("  -m    Re-assign the Maximus lastread pointers in USER.BBS.\n");

  exit(1);
}


struct tm * pascal OpusDate_to_TmDate(struct tm *tmdate,struct _stamp *opusdate)
{
  tmdate->tm_mday=opusdate->date.da;
  tmdate->tm_mon =opusdate->date.mo-1;
  tmdate->tm_year=opusdate->date.yr+80;

  tmdate->tm_hour=opusdate->time.hh;
  tmdate->tm_min=opusdate->time.mm;
  tmdate->tm_sec=opusdate->time.ss*2;

  return tmdate;
}



#ifdef __TURBOC__
long mktime(struct tm *lt)
{
  (void)lt;
  return 0L;
}
#endif


/* First, the polynomial itself and its table of feedback terms.  The  */
/* polynomial is                                                       */
/* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
/* Note that we take it "backwards" and put the highest-order term in  */
/* the lowest-order bit.  The X^32 term is "implied"; the LSB is the   */
/* X^31 term, etc.  The X^0 term (usually shown as "+1") results in    */
/* the MSB being 1.                                                    */

/* Note that the usual hardware shift register implementation, which   */
/* is what we're using (we're merely optimizing it by doing eight-bit  */
/* chunks at a time) shifts bits into the lowest-order term.  In our   */
/* implementation, that means shifting towards the right.  Why do we   */
/* do it this way?  Because the calculated CRC must be transmitted in  */
/* order from highest-order term to lowest-order term.  UARTs transmit */
/* characters in order from LSB to MSB.  By storing the CRC this way,  */
/* we hand it to the UART in the order low-byte to high-byte; the UART */
/* sends each low-bit to hight-bit; and the result is transmission bit */
/* by bit from highest- to lowest-order term without requiring any bit */
/* shuffling on our part.  Reception works similarly.                  */

/* The feedback terms table consists of 256, 32-bit entries.  Notes:   */
/*                                                                     */
/*     The table can be generated at runtime if desired; code to do so */
/*     is shown later.  It might not be obvious, but the feedback      */
/*     terms simply represent the results of eight shift/xor opera-    */
/*     tions for all combinations of data and CRC register values.     */
/*                                                                     */
/*     The values must be right-shifted by eight bits by the "updcrc"  */
/*     logic; the shift must be unsigned (bring in zeroes).  On some   */
/*     hardware you could probably optimize the shift in assembler by  */
/*     using byte-swap instructions.                                   */

long cr3tab[] = {                         /* CRC polynomial 0xedb88320 */
0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
0x2d02ef8dL };

#define Z_32UpdateCRC(c,crc) (cr3tab[((int) crc ^ c) & 0xff] ^ ((crc >> 8) & 0x00FFFFFFL))

long CRC(char *s)
{
  long crc;
  char c;

  for (crc=-1L;*s;)
  {
    c=toupper(*s++);
    crc=Z_32UpdateCRC(c,crc);
  }

  return crc;
}


